@@ -16,6 +16,8 @@
  * along with this program.  If not, see <http://www.gnu.org/licenses/>.
  */
 
+#include "config.h"
+
 #include <Python.h>
 #include <structmember.h>
 
@@ -24,8 +26,8 @@
 #define PYTHON_MODULE_NAME  "pypamtest"
 
 #ifndef discard_const_p
-#if defined(__intptr_t_defined) || defined(HAVE_INTPTR_T)
-# define discard_const_p(type, ptr) ((type *)((intptr_t)(ptr)))
+#if defined(__intptr_t_defined) || defined(HAVE_UINTPTR_T)
+# define discard_const_p(type, ptr) ((type *)((uintptr_t)(ptr)))
 #else
 # define discard_const_p(type, ptr) ((type *)(ptr))
 #endif
@@ -67,9 +69,9 @@ static PyObject *PyExc_PamTestError;
  *** helper functions
  **********************************************************/
 
-static const char *repr_fmt = "{ pam_operation [%d] "
-			      "expected_rv [%d] "
-			      "flags [%d] }";
+#define REPR_FMT "{ pam_operation [%d] " \
+			      "expected_rv [%d] " \
+			      "flags [%d] }"
 
 static char *py_strdup(const char *string)
 {
@@ -133,9 +135,18 @@ static void free_string_list(char **list)
 	PyMem_Free(list);
 }
 
-static char **new_conv_list(const int list_size)
+static char **new_conv_list(const size_t list_size)
 {
 	char **list;
+	size_t i;
+
+	if (list_size == 0) {
+		return NULL;
+	}
+
+	if (list_size + 1 < list_size) {
+		return NULL;
+	}
 
 	list = PyMem_New(char *, list_size + 1);
 	if (list == NULL) {
@@ -143,7 +154,7 @@ static char **new_conv_list(const int list_size)
 	}
 	list[list_size] = NULL;
 
-	for (int i =0; i < list_size; i++) {
+	for (i = 0; i < list_size; i++) {
 		list[i] = PyMem_New(char, PAM_MAX_MSG_SIZE);
 		if (list[i] == NULL) {
 			PyMem_Free(list);
@@ -155,11 +166,13 @@ static char **new_conv_list(const int list_size)
 	return list;
 }
 
-static const char **sequence_as_string_list(PyObject *seq,
-					    const char *paramname)
+static int sequence_as_string_list(PyObject *seq,
+				   const char *paramname,
+				   const char **str_list[],
+				   size_t *num_str_list)
 {
 	const char *p = paramname ? paramname : "attribute values";
-	const char **ret;
+	const char **result;
 	PyObject *utf_item;
 	int i;
 	Py_ssize_t len;
@@ -168,18 +181,18 @@ static const char **sequence_as_string_list(PyObject *seq,
 	if (!PySequence_Check(seq)) {
 		PyErr_Format(PyExc_TypeError,
 			     "The object must be a sequence\n");
-		return NULL;
+		return -1;
 	}
 
 	len = PySequence_Size(seq);
 	if (len == -1) {
-		return NULL;
+		return -1;
 	}
 
-	ret = PyMem_New(const char *, (len + 1));
-	if (!ret) {
+	result = PyMem_New(const char *, (len + 1));
+	if (result == NULL) {
 		PyErr_NoMemory();
-		return NULL;
+		return -1;
 	}
 
 	for (i = 0; i < len; i++) {
@@ -191,20 +204,24 @@ static const char **sequence_as_string_list(PyObject *seq,
 		utf_item = get_utf8_string(item, p);
 		if (utf_item == NULL) {
 			Py_DECREF(item);
-			return NULL;
+			return -1;
 		}
 
-		ret[i] = py_strdup(PyBytes_AsString(utf_item));
+		result[i] = py_strdup(PyBytes_AsString(utf_item));
 		Py_DECREF(utf_item);
-		if (!ret[i]) {
+		if (result[i] == NULL) {
 			Py_DECREF(item);
-			return NULL;
+			return -1;
 		}
 		Py_DECREF(item);
 	}
 
-	ret[i] = NULL;
-	return ret;
+	result[i] = NULL;
+
+	*str_list = result;
+	*num_str_list = (size_t)len;
+
+	return 0;
 }
 
 static PyObject *string_list_as_tuple(char **str_list)
@@ -214,7 +231,7 @@ static PyObject *string_list_as_tuple(char **str_list)
 	PyObject *tup;
 	PyObject *py_str;
 
-	for (len=0; len < PAM_CONV_MSG_MAX; len++) {
+	for (len=0; str_list[len] != NULL; len++) {
 		if (str_list[len][0] == '\0') {
 			/* unused string, stop counting */
 			break;
@@ -258,7 +275,7 @@ set_pypamtest_exception(PyObject *exc,
 			size_t num_tests)
 {
 	PyObject *obj = NULL;
-	/* repr_fmt is fixed and contains just %d expansions, so this is safe */
+	/* REPR_FMT contains just %d expansions, so this is safe */
 	char test_repr[256] = { '\0' };
 	union {
 		char *str;
@@ -277,7 +294,7 @@ set_pypamtest_exception(PyObject *exc,
 	if (perr == PAMTEST_ERR_CASE) {
 		failed = _pamtest_failed_case(tests, num_tests);
 		if (failed) {
-			snprintf(test_repr, sizeof(test_repr), repr_fmt,
+			snprintf(test_repr, sizeof(test_repr), REPR_FMT,
 				 failed->pam_operation,
 				 failed->expected_rv,
 				 failed->flags);
@@ -413,7 +430,7 @@ static int TestCase_init(TestCaseObject *self,
  */
 static PyObject *TestCase_repr(TestCaseObject *self)
 {
-	return PyUnicode_FromFormat(repr_fmt,
+	return PyUnicode_FromFormat(REPR_FMT,
 				    self->pam_operation,
 				    self->expected_rv,
 				    self->flags);
@@ -797,31 +814,46 @@ static int fill_conv_data(PyObject *py_echo_off,
 			  PyObject *py_echo_on,
 			  struct pamtest_conv_data *conv_data)
 {
+	size_t conv_count = 0;
+	size_t count = 0;
+	int rc;
+
 	conv_data->in_echo_on = NULL;
 	conv_data->in_echo_off = NULL;
 	conv_data->out_err = NULL;
 	conv_data->out_info = NULL;
 
 	if (py_echo_off != NULL) {
-		conv_data->in_echo_off = sequence_as_string_list(py_echo_off,
-								 "echo_off");
-		if (conv_data->in_echo_off == NULL) {
+		rc = sequence_as_string_list(py_echo_off,
+					     "echo_off",
+					     &conv_data->in_echo_off,
+					     &count);
+		if (rc != 0) {
 			free_conv_data(conv_data);
 			return ENOMEM;
 		}
+		conv_count += count;
 	}
 
 	if (py_echo_on != NULL) {
-		conv_data->in_echo_on = sequence_as_string_list(py_echo_on,
-								"echo_on");
-		if (conv_data->in_echo_on == NULL) {
+		rc = sequence_as_string_list(py_echo_on,
+					     "echo_on",
+					     &conv_data->in_echo_on,
+					     &count);
+		if (rc != 0) {
 			free_conv_data(conv_data);
 			return ENOMEM;
 		}
+		conv_count += count;
 	}
 
-	conv_data->out_info = new_conv_list(PAM_CONV_MSG_MAX);
-	conv_data->out_err = new_conv_list(PAM_CONV_MSG_MAX);
+	if (conv_count > PAM_CONV_MSG_MAX) {
+		free_conv_data(conv_data);
+		return ENOMEM;
+	}
+
+	conv_data->out_info = new_conv_list(conv_count);
+	conv_data->out_err = new_conv_list(conv_count);
 	if (conv_data->out_info == NULL || conv_data->out_err == NULL) {
 		free_conv_data(conv_data);
 		return ENOMEM;
@@ -995,12 +1027,14 @@ static struct PyModuleDef pypamtestdef = {
  *** Initialize the module
  **********************************************************/
 
+#if PY_VERSION_HEX >= 0x02070000 /* >= 2.7.0 */
 PyDoc_STRVAR(PamTestError__doc__,
 "pypamtest specific exception\n\n"
 "This exception is raised if the _pamtest() function fails. If _pamtest() "
 "returns PAMTEST_ERR_CASE (a test case returns unexpected error code), then "
 "the exception also details which test case failed."
 );
+#endif
 
 #if IS_PYTHON3
 PyMODINIT_FUNC PyInit_pypamtest(void)
@@ -1025,10 +1059,17 @@ PyMODINIT_FUNC initpypamtest(void)
 			  pypamtest_module_methods);
 #endif
 
+#if PY_VERSION_HEX >= 0x02070000 /* >= 2.7.0 */
 	PyExc_PamTestError = PyErr_NewExceptionWithDoc(discard_const_p(char, "pypamtest.PamTestError"),
 						       PamTestError__doc__,
 						       PyExc_EnvironmentError,
 						       NULL);
+#else /* < 2.7.0 */
+	PyExc_PamTestError = PyErr_NewException(discard_const_p(char, "pypamtest.PamTestError"),
+						       PyExc_EnvironmentError,
+						       NULL);
+#endif
+
 	if (PyExc_PamTestError == NULL) {
 		RETURN_ON_ERROR;
 	}
