@@ -127,7 +127,7 @@ static int setup_passdb(void **state)
 	db = getcwd(passdb_path, PATH_MAX);
 	assert_non_null(db);
 	assert_true(strlen(passdb_path) + sizeof("/passdb") < PATH_MAX);
-	db = strncat(passdb_path, "/passdb", sizeof("/passdb"));
+	db = strncat(passdb_path, "/passdb", strlen("/passdb") + 1);
 
 	rv = setenv("PAM_MATRIX_PASSWD", passdb_path, 1);
 	assert_int_equal(rv, 0);
